# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from future.utils import iteritems
from mssqlscripter.jsonrpc.contracts import Request

import copy
import logging

logger = logging.getLogger(u'mssqlscripter.jsonrpc.contracts.scriptingservice')


class ScriptingRequest(Request):
    """
        SqlTools Service scripting service scripting request.
    """
    METHOD_NAME = u'scripting/script'

    def __init__(self, id, json_rpc_client, parameters):
        """
            Create a scripting request command.
        """
        assert id != 0
        self.id = id
        self.finished = False
        self.json_rpc_client = json_rpc_client
        self.params = ScriptingParams(parameters)
        self.decoder = ScriptingResponseDecoder()

    def execute(self):
        """
            submit scripting request to sql tools service.
        """
        logger.info(
            u'Submitting scripting request id: {} with targetfile: {}'.format(
                self.id, self.params.file_path))

        scrubbed_parameters = copy.deepcopy(self.params)
        scrubbed_parameters.connection_string = '*********'
        logger.debug(scrubbed_parameters.format())
        self.json_rpc_client.submit_request(
            self.METHOD_NAME, self.params.format(), self.id)

    def get_response(self):
        """
            Get latest response, event or exception if it occured.
        """
        try:
            response = self.json_rpc_client.get_response(self.id)
            decoded_response = None

            if response:
                logger.debug(response)
                # Decode response to either response or event type.
                decoded_response = self.decoder.decode_response(response)

                logger.debug(
                    u'Scripting request received response: {}'.format(decoded_response))
                if (isinstance(decoded_response, ScriptCompleteEvent)):
                    self.finished = True
                    self.json_rpc_client.request_finished(self.id)

            return decoded_response

        except Exception as error:
            # Return a scripting error event.
            self.finished = True
            self.json_rpc_client.request_finished(self.id)
            logger.debug('Scripting request received exception: {}'.format(str(error)))
            exception = {
                u'operationId': self.id,
                u'sequenceNumber': None,
                u'success': False,
                u'canceled': False,
                u'hasError': True,
                u'errorMessage': u'Scripting request encountered a exception',
                u'errorDetails': error.args,
            }

            return ScriptCompleteEvent(exception)

    def completed(self):
        """
            Get current request state.
        """
        return self.finished


class ScriptingParams(object):
    """
        Scripting request parameters.
    """

    def __init__(self, parameters):
        self.file_path = parameters[u'FilePath']
        self.connection_string = parameters[u'ConnectionString']
        self.script_destination = parameters[u'ScriptDestination']
        self.scripting_options = ScriptingOptions(parameters)

        self.include_schema = parameters[u'IncludeSchemas'] if u'IncludeSchemas' in parameters else None
        self.exclude_schema = parameters[u'ExcludeSchemas'] if u'ExcludeSchemas' in parameters else None
        self.include_type = parameters[u'IncludeTypes'] if u'IncludeTypes' in parameters else None
        self.exclude_type = parameters[u'ExcludeTypes'] if u'ExcludeTypes' in parameters else None

        # List of scripting objects.
        self.include_objects = ScriptingObjects(
            parameters[u'IncludeObjects'] if u'IncludeObjects' in parameters else None)
        self.exclude_objects = ScriptingObjects(
            parameters[u'ExcludeObjects'] if u'ExcludeObjects' in parameters else None)

    def format(self):
        """
            Format paramaters into a dictionary.
        """
        return {u'FilePath': self.file_path,
                u'ConnectionString': self.connection_string,
                u'IncludeObjectCriteria': self.include_objects.format(),
                u'ExcludeObjectCriteria': self.exclude_objects.format(),
                u'IncludeSchemas': self.include_schema,
                u'ExcludeSchemas': self.exclude_schema,
                u'IncludeTypes': self.include_type,
                u'ExcludeTypes': self.exclude_type,
                u'ScriptOptions': self.scripting_options.get_options(),
                u'ScriptDestination': self.script_destination}


class ScriptingObjects(object):
    """
        Represent a database object via it's type, schema, and name.
    """

    def __init__(self, scripting_objects):
        self.list_of_objects = []
        if scripting_objects:
            for item in scripting_objects:
                index = item.find('.')
                if index > 0:
                    schema = item[0:index]
                    name = item[index + 1:]
                else:
                    schema = None
                    name = item
                self.add_scripting_object(schema=schema, name=name)

    def add_scripting_object(self, script_type=None, schema=None, name=None):
        """
            Serialize scripting object into a JSON Scripting object.
        """
        object_dict = {
            u'Type': script_type,
            u'Schema': schema,
            u'Name': name
        }

        self.list_of_objects.append(object_dict)

    def format(self):
        return self.list_of_objects


class ScriptingOptions(object):
    """
        Advanced scripting options.
    """
    scripting_option_map = {
        u'TypeOfDataToScript': [
            u'SchemaAndData',
            u'DataOnly',
            u'SchemaOnly'],
        u'ScriptCreateDrop': [
            u'ScriptCreate',
            u'ScriptDrop',
            u'ScriptCreateDrop'],
        u'TargetDatabaseEngineType': [
            u'SingleInstance',
            u'SqlAzure'],
        u'ScriptStatistics': [
            u'ScriptStatsAll',
            u'ScriptStatsNone',
            u'ScriptStatsDll'],
        u'ScriptCompatibilityOption': [
            u'Script90Compat',
            u'Script100Compat',
            u'Script105Compat',
            u'Script110Compat',
            u'Script120Compat',
            u'Script130Compat',
            u'Script140Compat'],
        u'TargetDatabaseEngineEdition': [
            u'SqlServerStandardEdition',
            u'SqlServerPersonalEdition',
            u'SqlServerExpressEdition',
            u'SqlServerEnterpriseEdition',
            u'SqlServerStretchDatabaseEdition',
            u'SqlAzureDatabaseEdition',
            u'SqlDatawarehouseEdition']}

    def __init__(self, parameters=None):
        """
            Create default or non default scripting options based on parameters.
        """
        # General Default scripting options.
        self.ScriptAnsiPadding = False
        self.AppendToFile = False
        self.IncludeIfNotExists = False
        self.ContinueScriptingOnError = False
        self.ConvertUDDTToBaseType = False
        self.GenerateScriptForDependentObjects = False
        self.IncludeDescriptiveHeaders = False
        self.IncludeSystemConstraintNames = False
        self.IncludeUnsupportedStatements = False
        self.SchemaQualify = False
        self.Bindings = False
        self.Collation = False
        self.Default = False
        self.ScriptExtendedProperties = False
        self.ScriptLogins = False
        self.ScriptObjectLevelPermissions = False
        self.ScriptOwner = False
        self.ScriptUseDatabase = False

        # Default Table/View options.
        self.ScriptChangeTracking = False
        self.ScriptCheckConstraints = False
        self.ScriptDataCompressionOptions = False
        self.ScriptForeignKeys = False
        self.ScriptFullTextIndexes = False
        self.ScriptIndexes = False
        self.ScriptPrimaryKeys = False
        self.ScriptTriggers = False
        self.UniqueKeys = False

        # Scripting options that are limited.
        self.TypeOfDataToScript = u'SchemaOnly'
        self.ScriptCreateDrop = u'ScriptCreate'
        self.TargetDatabaseEngineType = u'SingleInstance'
        self.ScriptStatistics = u'ScriptStatsNone'
        self.ScriptCompatibilityOption = u'Script140Compat'
        self.TargetDatabaseEngineEdition = u'SqlServerStandardEdition'

        if parameters:
            self.update_options(parameters)

    def update_options(self, parameters):
        """
            Update default options to passed in options.
        """
        default_options = vars(self)
        for option, value in iteritems(parameters):
            if option in default_options:
                if option in self.scripting_option_map:
                    if value not in self.scripting_option_map[option]:
                        raise ValueError(
                            u'Option: {} has invalid value: {}'.format(
                                option, value))
                elif not isinstance(value, bool):
                    raise ValueError(
                        u'Option: {} has unexpected value type" {}'.format(
                            option, value))
                # set the value if we pass all the checks.
                default_options[option] = value

    def get_options(self):
        """
            Return a dictionary of all options
        """
        return vars(self)

#
#   Various Scripting Events.
#


class ScriptCompleteEvent(object):
    def __init__(self, params):
        self.operation_id = params[u'operationId']
        self.sequenceNumber = params[u'sequenceNumber']
        self.error_details = params[u'errorDetails']
        self.error_message = params[u'errorMessage']
        self.has_error = params[u'hasError']
        self.canceled = params[u'canceled']
        self.success = params[u'success']


class ScriptPlanNotificationEvent(object):
    def __init__(self, params):
        self.operation_id = params[u'operationId']
        self.sequenceNumber = params[u'sequenceNumber']
        self.scripting_objects = params[u'scriptingObjects']
        self.count = params[u'count']


class ScriptProgressNotificationEvent(object):
    def __init__(self, params):
        self.operation_id = params[u'operationId']
        self.sequenceNumber = params[u'sequenceNumber']
        self.scripting_object = params[u'scriptingObject']
        self.status = params[u'status']
        self.completed_count = params[u'completedCount']
        self.total_count = params[u'totalCount']


class ScriptResponse(object):
    def __init__(self, params):
        self.operation_id = params[u'operationId']


class ScriptingResponseDecoder(object):
    """
        Decode response dictionary into scripting parameter type.
    """

    def __init__(self):
        # response map.
        self.response_dispatcher = {
            u'scripting/scriptComplete': ScriptCompleteEvent,
            u'scripting/scriptPlanNotification': ScriptPlanNotificationEvent,
            u'scripting/scriptProgressNotification': ScriptProgressNotificationEvent,
            u'id': ScriptResponse}

    def decode_response(self, obj):
        """
            Decode valid response.
        """
        if u'method' in obj:
            response_name = obj[u'method']
            if response_name in self.response_dispatcher:
                # Handle event received.
                return self.response_dispatcher[response_name](obj[u'params'])

        if u'id' in obj and u'result' in obj:
            # Handle response received.
            return self.response_dispatcher[u'id'](obj[u'result'])

        logger.debug(
            u'Unable to decode response to a event type: {}'.format(obj))
        # Unable to decode, return json string.
        return obj



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import logging
import logging.handlers
import os


def get_config_log_dir():
    """
        Retrieve logging directory, create it if it doesn't exist.
    """
    log_dir = os.path.expanduser(os.path.join(u'~', u'.mssqlscripter'))
    if (not os.path.exists(log_dir)):
        os.makedirs(log_dir)
    return log_dir


def get_config_log_file():
    """
        Retrieve log file path.
    """
    return os.path.join(get_config_log_dir(), u'mssql-scripter.log')


def initialize_logger():
    """
        Initializes root logger for scripter with rotating file handler. Should only be called once from main program.
    """
    scripter_logger = logging.getLogger('mssqlscripter')
    scripter_logger.setLevel(logging.DEBUG)
    handler = logging.handlers.RotatingFileHandler(
        get_config_log_file(), maxBytes=2 * 1024 * 1024, backupCount=10)

    formatter = logging.Formatter(
        u'%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    handler.setFormatter(formatter)
    scripter_logger.addHandler(handler)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import sys

import mssqlscripter.main

try:
    args = sys.argv[1:]
    exit_code = mssqlscripter.main.main(args)
    sys.exit(exit_code)
except EnvironmentError as error:
    sys.stderr.write(str(error))
    sys.exit(1)
except KeyboardInterrupt as error:
    sys.exit(2)
except Exception as error:
    sys.stderr.write(str(error))
    sys.exit(3)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import copy
import io
import logging
import os
import platform
import subprocess
import sys
import tempfile
import time


import mssqlscripter.scripterlogging as scripterlogging
import mssqlscripter.argparser as parser
import mssqlscripter.scriptercallbacks as scriptercallbacks
import mssqlscripter.sqltoolsclient as sqltoolsclient
import mssqltoolsservice

logger = logging.getLogger(u'mssqlscripter.main')


def main(args):
    """
        Main entry point to mssql-scripter.
    """
    scripterlogging.initialize_logger()
    logger.info('Python Information :{}'.format(sys.version_info))
    logger.info('System Information: system={} architecture={} version={}'.format(platform.system(), platform.architecture()[0], platform.version()))

    parameters = parser.parse_arguments(args)
    scrubbed_parameters = copy.deepcopy(parameters)

    try:
        scrubbed_parameters.ConnectionString = '*******'
        scrubbed_parameters.Password = '********'
    except AttributeError:
        # Password was not given, using integrated auth.
        pass

    logger.info(scrubbed_parameters)

    temp_file_path = None

    if not parameters.FilePath and parameters.ScriptDestination == 'ToSingleFile':
        # Generate and track the temp file.
        temp_file_path = tempfile.NamedTemporaryFile(
            prefix=u'mssqlscripter_', delete=False).name
        parameters.FilePath = temp_file_path

    sqltoolsservice_args = [mssqltoolsservice.get_executable_path()]

    if parameters.EnableLogging:
        sqltoolsservice_args.append('--enable-logging')
        sqltoolsservice_args.append('--log-dir')
        sqltoolsservice_args.append(scripterlogging.get_config_log_dir())

    logger.debug('Loading mssqltoolsservice with arguments {}'.format(sqltoolsservice_args))
    try:
        # Start mssqltoolsservice program.
        tools_service_process = subprocess.Popen(
            sqltoolsservice_args,
            bufsize=0,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE)

        # Python 2.7 uses the built-in File type when referencing the subprocess.PIPE.
        # This built-in type for that version blocks on readinto() because it attempts to fill buffer.
        # Wrap a FileIO around it to use a different implementation that does not attempt to fill the buffer
        # on readinto().
        std_out_wrapped = io.open(
            tools_service_process.stdout.fileno(),
            u'rb',
            buffering=0,
            closefd=False)

        sql_tools_client = sqltoolsclient.SqlToolsClient(
            tools_service_process.stdin,
            std_out_wrapped)

        scripting_request = sql_tools_client.create_request(
            u'scripting_request', vars(parameters))
        scripting_request.execute()

        while not scripting_request.completed():
            response = scripting_request.get_response()
            if response:
                scriptercallbacks.handle_response(response, parameters.DisplayProgress)
            else:
                # The sleep prevents burning up the CPU and lets other threads get scheduled.
                time.sleep(0.1)

        # Only write to stdout if user did not provide a file path.
        logger.info('stdout current encoding: {}'.format(sys.stdout.encoding))
        if temp_file_path:
            with io.open(parameters.FilePath, encoding=u'utf-8') as script_file:
                for line in script_file.readlines():
                    sys.stdout.write(line)

    finally:

        sql_tools_client.shutdown()
        tools_service_process.kill()
        # 1 second time out, allow tools service process to be killed.
        time.sleep(1)
        # None value indicates process has not terminated.
        if not tools_service_process.poll():
            sys.stderr.write(
                u'Sql Tools Service process was not shut down properly.')
        try:
            # Remove the temp file if we generated one.
            if temp_file_path:
                os.remove(temp_file_path)
        except Exception:
            # Suppress exceptions.
            pass


if __name__ == u'__main__':
    main(sys.argv[1:])



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import argparse
import getpass
import mssqlscripter
import os
import shutil
import sys

MSSQL_SCRIPTER_CONNECTION_STRING = u'MSSQL_SCRIPTER_CONNECTION_STRING'
MSSQL_SCRIPTER_PASSWORD = u'MSSQL_SCRIPTER_PASSWORD'


def parse_arguments(args):
    """
        Initialize parser with scripter options.
    """
    parser = argparse.ArgumentParser(
        prog=u'mssql-scripter',
        description=u'Microsoft SQL Server Scripter Command Line Tool. ' +
        'Version {}'.format(mssqlscripter.__version__))

    group_connection_options = parser.add_mutually_exclusive_group()
    group_connection_options.add_argument(
        u'--connection-string',
        dest=u'ConnectionString',
        metavar=u'',
        help=u'Connection string of database to script. If connection string and server are not supplied, defaults to value in environment variable MSSQL_SCRIPTER_CONNECTION_STRING.')
    group_connection_options.add_argument(
        u'-S', u'--server',
        dest=u'Server',
        metavar=u'',
        help=u'Server name.')

    parser.add_argument(
        u'-d', u'--database',
        dest=u'Database',
        metavar=u'',
        help=u'Database name.')

    parser.add_argument(
        u'-U', u'--user',
        dest=u'UserId',
        metavar=u'',
        help=u'Login ID for server.')

    parser.add_argument(
        u'-P', u'--password',
        dest=u'Password',
        metavar=u'',
        help=u'If not supplied, defaults to value in environment variable MSSQL_SCRIPTER_PASSWORD.')

    # Basic parameters.
    parser.add_argument(
        u'-f', u'--file-path',
        dest=u'FilePath',
        metavar=u'',
        default=None,
        help=u'File to script out to or directory name if scripting file per object.')

    parser.add_argument(
        u'--file-per-object',
        dest=u'ScriptDestination',
        action=u'store_const',
        const=u'ToFilePerObject',
        default=u'ToSingleFile',
        help=u'By default script to a single file. If supplied and given a directory for --file-path, script a file per object to that directory.')

    group_type_of_data = parser.add_mutually_exclusive_group()
    group_type_of_data.add_argument(
        u'--data-only',
        dest=u'TypeOfDataToScript',
        action=u'store_const',
        const=u'DataOnly',
        default=u'SchemaOnly',
        help=u'By default only the schema is scripted. if supplied, generate scripts that contains data only.')
    group_type_of_data.add_argument(
        u'--schema-and-data',
        dest=u'TypeOfDataToScript',
        action=u'store_const',
        const=u'SchemaAndData',
        default=u'SchemaOnly',
        help=u'By default only the schema is scripted. if supplied, generate scripts that contain schema and data.')

    group_create_drop = parser.add_mutually_exclusive_group()
    group_create_drop.add_argument(
        u'--script-create',
        dest=u'ScriptCreateDrop',
        action=u'store_const',
        const=u'ScriptCreate',
        default=u'ScriptCreate',
        help=u'Script object CREATE statements.')
    group_create_drop.add_argument(
        u'--script-drop',
        dest=u'ScriptCreateDrop',
        action=u'store_const',
        const=u'ScriptDrop',
        default=u'ScriptCreate',
        help=u'Script object DROP statements.')
    group_create_drop.add_argument(
        u'--script-drop-create',
        dest=u'ScriptCreateDrop',
        action=u'store_const',
        const=u'ScriptCreateDrop',
        default=u'ScriptCreate',
        help=u'Script object CREATE and DROP statements.')

    parser.add_argument(
        u'--target-server-version',
        dest=u'ScriptCompatibilityOption',
        choices=[
            '2005',
            '2008',
            '2008R2',
            '2012',
            '2014',
            '2016',
            'vNext',
            'AzureDB',
            'AzureDW'],
        default=u'2016',
        help=u'Script only features compatible with the specified SQL Version.')

    parser.add_argument(
        u'--target-server-edition',
        dest=u'TargetDatabaseEngineEdition',
        choices=[
            u'Standard',
            u'Personal',
            u'Express',
            u'Enterprise',
            u'Stretch'],
        default=u'Enterprise',
        help=u'Script only features compatible with the specified SQL Server database edition.')

    parser.add_argument(
        u'--include-objects',
        dest=u'IncludeObjects',
        nargs=u'*',
        type=str,
        metavar=u'',
        help=u'Database objects to include in script.')

    parser.add_argument(
        u'--exclude-objects',
        dest=u'ExcludeObjects',
        nargs=u'*',
        type=str,
        metavar=u'',
        help=u'Database objects to exclude from script.')

    parser.add_argument(
        u'--include-schemas',
        dest=u'IncludeSchemas',
        nargs=u'*',
        type=str,
        metavar=u'',
        help=u'Database objects of this schema to include in script.')

    parser.add_argument(
        u'--exclude-schemas',
        dest=u'ExcludeSchemas',
        nargs=u'*',
        type=str,
        metavar=u'',
        help=u'Database objects of this schema to exclude from script.')

    parser.add_argument(
        u'--include-types',
        dest=u'IncludeTypes',
        nargs=u'*',
        type=str,
        metavar=u'',
        help=u'Database objects of this type to include in script.')

    parser.add_argument(
        u'--exclude-types',
        dest=u'ExcludeTypes',
        nargs=u'*',
        type=str,
        metavar=u'',
        help=u'Database objects of this type to exclude from script.')

    # General boolean Scripting Options
    parser.add_argument(
        u'--ansi-padding',
        dest=u'ScriptAnsiPadding',
        action=u'store_true',
        default=False,
        help=u'Generates ANSI Padding statements.')

    parser.add_argument(
        u'--append',
        dest=u'AppendToFile',
        action=u'store_true',
        default=False,
        help=u'Append script to file.')

    parser.add_argument(
        u'--check-for-existence',
        dest=u'IncludeIfNotExists',
        action=u'store_true',
        default=False,
        help=u'Check that an object with the given name exists before dropping or altering or that an object with the given name does not exist before creating.')

    parser.add_argument(
        u'-r',
        u'--continue-on-error',
        dest=u'ContinueScriptingOnError',
        action=u'store_true',
        default=False,
        help=u'Continue scripting on error.')

    parser.add_argument(
        u'--convert-uddts',
        dest=u'ConvertUDDTToBaseType',
        action=u'store_true',
        default=False,
        help=u'Convert user-defined data types to base types.')

    parser.add_argument(
        u'--include-dependencies',
        dest=u'GenerateScriptForDependentObjects',
        action=u'store_true',
        default=False,
        help=u'Generate script for the dependent objects for each object scripted.')

    parser.add_argument(
        u'--exclude-headers',
        dest=u'IncludeDescriptiveHeaders',
        action=u'store_false',
        default=True,
        help=u'Exclude descriptive headers for each object scripted.')

    parser.add_argument(
        u'--constraint-names',
        dest=u'IncludeSystemConstraintNames',
        action=u'store_true',
        default=False,
        help=u'Include system constraint names to enforce declarative referential integrity.')

    parser.add_argument(
        u'--unsupported-statements',
        dest=u'IncludeUnsupportedStatements',
        action=u'store_true',
        default=False,
        help=u'Include statements in the script that are not supported on the target SQL Server Version.')

    parser.add_argument(
        u'--disable-schema-qualification',
        dest=u'SchemaQualify',
        action=u'store_false',
        default=True,
        help=u'Do not prefix object names with the object schema.')

    parser.add_argument(
        u'--bindings',
        dest=u'Bindings',
        action=u'store_true',
        default=False,
        help=u'Script options to set binding options.')

    parser.add_argument(
        u'--collation',
        dest=u'Collation',
        action=u'store_true',
        default=False,
        help=u'Script the objects that use collation.')

    parser.add_argument(
        u'--exclude-defaults',
        dest=u'Default',
        action=u'store_false',
        default=True,
        help=u'Do not script the default values.')

    parser.add_argument(
        u'--exclude-extended-properties',
        dest=u'ScriptExtendedProperties',
        action=u'store_false',
        default=True,
        help=u'Exclude extended properties for each object scripted.')

    parser.add_argument(
        u'--logins',
        dest=u'ScriptLogins',
        action=u'store_true',
        default=False,
        help=u'Script all logins available on the server, passwords will not be scripted.')

    parser.add_argument(
        u'--object-permissions',
        dest=u'ScriptObjectLevelPermissions',
        action=u'store_true',
        default=False,
        help=u'Generate object-level permissions.')

    parser.add_argument(
        u'--owner',
        dest=u'ScriptOwner',
        action=u'store_true',
        default=False,
        help=u'Script owner for the objects.')

    parser.add_argument(
        u'--exclude-use-database',
        dest=u'ScriptUseDatabase',
        action=u'store_false',
        default=True,
        help=u'Do not generate USE DATABASE statement.')

    parser.add_argument(
        u'--statistics',
        dest=u'ScriptStatistics',
        action=u'store_const',
        const=u'ScriptStatsAll',
        default=u'ScriptStatsNone',
        help=u'Script all statistics.')

    parser.add_argument(
        u'--database-engine-type',
        dest=u'TargetDatabaseEngineType',
        # This parameter is determined based on engine edition and version in
        # the background. User cannot select it.
        action=u'store_const',
        const=u'SingleInstance',
        default=u'SingleInstance',
        help=argparse.SUPPRESS)

    # Table/View Options
    parser.add_argument(
        u'--change-tracking',
        dest=u'ScriptChangeTracking',
        action=u'store_true',
        default=False,
        help=u'Script the change tracking information.')

    parser.add_argument(
        u'--exclude-check-constraints',
        dest=u'ScriptCheckConstraints',
        action=u'store_false',
        default=True,
        help=u'Exclude check constraints for each table or view scripted.')

    parser.add_argument(
        u'--data-compressions',
        dest=u'ScriptDataCompressionOptions',
        action=u'store_true',
        default=False,
        help=u'Script the data compression information.')

    parser.add_argument(
        u'--exclude-foreign-keys',
        dest=u'ScriptForeignKeys',
        action=u'store_false',
        default=True,
        help=u'Exclude foreign keys for each table scripted.')

    parser.add_argument(
        u'--exclude-full-text-indexes',
        dest=u'ScriptFullTextIndexes',
        action=u'store_false',
        default=True,
        help=u'Exclude full-text indexes for each table or indexed view scripted.')

    parser.add_argument(
        u'--exclude-indexes',
        dest=u'ScriptIndexes',
        action=u'store_false',
        default=True,
        help=u'Exclude indexes (XML and clustered) for each table or indexed view scripted.')

    parser.add_argument(
        u'--exclude-primary-keys',
        dest=u'ScriptPrimaryKeys',
        action=u'store_false',
        default=True,
        help=u'Exclude primary keys for each table or view scripted.')

    parser.add_argument(
        u'--exclude-triggers',
        dest=u'ScriptTriggers',
        action=u'store_false',
        default=True,
        help=u'Exclude triggers for each table or view scripted.')

    parser.add_argument(
        u'--exclude-unique-keys',
        dest=u'UniqueKeys',
        action=u'store_false',
        default=True,
        help=u'Exclude unique keys for each table or view scripted.')

    # Configuration Options.
    parser.add_argument(
        u'--display-progress',
        dest=u'DisplayProgress',
        action=u'store_true',
        default=False,
        help=u'Display scripting progress.')

    parser.add_argument(
        u'--enable-toolsservice-logging',
        dest=u'EnableLogging',
        action=u'store_true',
        default=False,
        help=u'Enable verbose logging.')

    parser.add_argument(
        u'--version',
        action=u'version',
        version='{}'.format(mssqlscripter.__version__))

    parameters = parser.parse_args(args)
    verify_directory(parameters)

    if parameters.Server:
        build_connection_string(parameters)
    elif parameters.ConnectionString is None:
        # Check environment variable for connection string.
        if not get_connection_string_from_environment(parameters):
            sys.stdout.write(u'Please specify connection information using --connection-string or --server and/or --database --user.\n')
            sys.exit()

    map_server_options(parameters)
    return parameters


def verify_directory(parameters):
    """
        If creating a file per object, create the directory if it does not exist.
    """

    target_directory = parameters.FilePath
    if parameters.ScriptDestination == 'ToFilePerObject':
        if not os.path.exists(target_directory):
            os.makedirs(target_directory)
        # Give warning to user that target directory was not empty.
        if os.listdir(target_directory):
            sys.stdout.write(u'warning: Target directory {} was not empty.'.format(target_directory))


def get_connection_string_from_environment(parameters):
    """
        Get connection string from environment variable.
    """
    if MSSQL_SCRIPTER_CONNECTION_STRING in os.environ:
        parameters.ConnectionString = os.environ[MSSQL_SCRIPTER_CONNECTION_STRING]
        return True

    return False


def build_connection_string(parameters):
    """
        Build connection string.
    """
    connection_string = u'Server={};'.format(parameters.Server)
    if parameters.Database:
        connection_string += u'Database={};'.format(parameters.Database)

    # Standard connection if user id is supplied.
    if parameters.UserId:
        connection_string += u'User Id={};'.format(parameters.UserId)
        # If no password supplied, check for environment variable.
        if parameters.Password is None and MSSQL_SCRIPTER_PASSWORD in os.environ:
            parameters.Password = os.environ[MSSQL_SCRIPTER_PASSWORD]

        connection_string += u'Password={};'.format(parameters.Password or getpass.getpass())

    else:
        connection_string += u'Integrated Security=True;'

    parameters.ConnectionString = connection_string


def map_server_options(parameters):
    """
        Map short form to long form name and maps Azure versions to their appropriate editions.
    """
    azure_server_edition_map = {
        u'AzureDB': u'SqlAzureDatabaseEdition',
        u'AzureDW': u'SqlDatawarehouseEdition',
    }

    on_prem_server_edition_map = {
        u'Standard': u'SqlServerStandardEdition',
        u'Personal': u'SqlServerPersonalEdition',
        u'Express': u'SqlServerExpressEdition',
        u'Enterprise': u'SqlServerEnterpriseEdition',
        u'Stretch': u'SqlServerStretchDatabaseEdition',
    }

    on_prem_server_version_map = {
        u'2005': u'Script90Compat',
        u'2008': u'Script100Compat',
        u'2008R2': u'Script105Compat',
        u'2012': u'Script110Compat',
        u'2014': u'Script120Compat',
        u'2016': u'Script130Compat',
        u'vNext': u'Script140Compat',
    }

    target_server_version = parameters.ScriptCompatibilityOption
    target_server_edition = parameters.TargetDatabaseEngineEdition
    # When targetting Azure, only the edition matters.
    if u'Azure' in target_server_version:
        # SMO requires 120 compat or higher when scripting Azure or AzureDW.
        parameters.ScriptCompatibilityOption = u'Script130Compat'
        parameters.TargetDatabaseEngineEdition = azure_server_edition_map[
            target_server_version]
        parameters.TargetDatabaseEngineType = u'SqlAzure'

    else:
        parameters.ScriptCompatibilityOption = on_prem_server_version_map[target_server_version]
        parameters.TargetDatabaseEngineEdition = on_prem_server_edition_map[
            target_server_edition]
        parameters.TargetDatabaseEngineType = u'SingleInstance'

    return parameters



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
import abc

ABC = abc.ABCMeta('ABC', (object,), {})  # compatibile with Python 2 *and* 3.


class Request(ABC):
    """
        Abstract request class.
    """
    @abc.abstractmethod
    def execute(self):
        """
            Executes the request.
        """
        pass

    @abc.abstractmethod
    def get_response(self):
        """
            Retrieves expected response.
        """
        pass

    @abc.abstractmethod
    def completed(self):
        """
            Return state of request.
        """
        pass



#!/usr/bin/env python

# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import io
import os
import platform as _platform
import sys

from setuptools import setup

# This version number is in place in two places and must be in sync with
# mssqltoolsservice's version in setup.py.
MSSQLSCRIPTER_VERSION = '1.0.0a20'

# If we have the source, validate our setup version matches source version.
# This will prevent uploading releases with mismatched versions. This will
# also ensure mssqlscripter's version is in sync with mssqltoolsservice.
try:
    with io.open('mssqlscripter/__init__.py', 'r', encoding='utf-8') as f:
        mssqlscripter_info = f.read()
    with io.open('mssqltoolsservice/mssqltoolsservice/__init__.py', 'r', encoding='utf-8') as f:
        mssqltoolsservice_info = f.read()
except IOError:
    pass
else:
    import re
    # Use regex to parse for version.
    scripter_version = re.search(
        r'__version__\s*=\s*[\'"](.+?)[\'"]',
        mssqlscripter_info)
    toolsservice_version = re.search(
        r'__version__\s*=\s*[\'"](.+?)[\'"]',
        mssqltoolsservice_info)

    if not scripter_version:
        print('Could not find __version__ in mssqlscripter/__init__.py')
        sys.exit(1)
    if not toolsservice_version:
        print(
            'Could not find __version__ in mssqltoolsservice/mssqltoolsservice/__init__.py')
        sys.exit(1)
    # Validate mssqlscripter source and setup versions.
    if scripter_version.group(1) != MSSQLSCRIPTER_VERSION:
        print('mssqlscripter version mismatch, source = "{}"; setup = "{}"'.format(
            scripter_version.group(1), MSSQLSCRIPTER_VERSION))
        sys.exit(1)
    # Validate mssqlscripter version with mssqltoolsservice.
    if scripter_version.group(1) != toolsservice_version.group(1):
        print(
            'mssqltoolsservice version mismatch, mssqscripter = "{}"; mssqltoolsservice = "{}"'.format(
                scripter_version.group(1),
                toolsservice_version.group(1)))
        sys.exit(1)

MSSQLTOOLSSERVICE_PACKAGE_NAME = 'mssqltoolsservice-{}=={}'
MSSQLTOOLSSERVICE_PACKAGE_SUFFIX = [
    'CentOS_7',
    'Debian_8',
    'Fedora_23',
    'openSUSE_13_2',
    'OSX_10_11_64',
    'RHEL_7',
    'Ubuntu_14',
    'Ubuntu_16',
    'Windows_7_64',
    'Windows_7_86'
]

LINUX_DISTRO_NO_VERSION = {
    'centos': 'CentOS_7',
    'ol': 'CentOS_7',
    'fedora': 'Fedora_23',
    'opensuse': 'OpenSUSE_13_2',
    'rhel': 'RHEL_7',
    'debian': 'Debian_8',
}

LINUX_DISTRO_WITH_VERSION = {
    'ubuntu':
        {
            '14': 'Ubuntu_14',
            '16': 'Ubuntu_16'
        },
    'elementary':
        {
            '0.3': 'Ubuntu_14',
            '0.4': 'Ubuntu_16'
        },
    'elementaryOS':
        {
            '0.3': 'Ubuntu_14',
            '0.4': 'Ubuntu_16'
        },
    'linuxmint':
        {
            '18': 'Ubuntu_16'
        },
    'galliumos':
        {
            '2.0': 'Ubuntu_16'
        },
}


def _get_runtime_id_helper(name, version):
    """
        Checks if linux distro name and version match to a supported package.
    """
    if name in LINUX_DISTRO_NO_VERSION:
        return LINUX_DISTRO_NO_VERSION[name]

    if name in LINUX_DISTRO_WITH_VERSION:
        for supported_version in LINUX_DISTRO_WITH_VERSION[name]:
            if version.startswith(supported_version):
                return LINUX_DISTRO_WITH_VERSION[name][supported_version]
    return None


def _get_linux_distro_runtime_id(content):
    """
        Parse content for linux distro run time id.
     """
    name = None
    version = None
    id_like = None

    # Try to find name, version and id_like best effort.
    for line in content.splitlines():
        key, value = line.rstrip().split('=')
        value = value.strip('"')
        if key == 'ID':
            name = value
        elif key == 'VERSION_ID':
            version = value
        elif key == 'ID_LIKE':
            id_like = value.split(' ')
        if name and version and id_like:
            break
    # First try the distribution name.
    run_time_id = _get_runtime_id_helper(name, version)

    # If we don't understand it, try the 'ID_LIKE' field.
    if not run_time_id and id_like:
        for name in id_like:
            run_time_id = _get_runtime_id_helper(name, version)
            if run_time_id:
                break

    return run_time_id


def _get_linux_distro_from_file():
    """
        Find linux distro based on
        https://www.freedesktop.org/software/systemd/man/os-release.html.
    """
    os_release_info_file = None

    if os.path.exists('/etc/os-release'):
        os_release_info_file = '/etc/os-release'
    elif os.path.exists('/usr/lib/os-release'):
        os_release_info_file = '/usr/lib/os-release'
    else:
        raise EnvironmentError('Error detecting Linux distro version.')

    with io.open(os_release_info_file, 'r', encoding='utf-8') as os_release_file:
        content = os_release_file.read()
        return _get_linux_distro_runtime_id(content)


def _get_runtime_id(
        system=_platform.system(),
        architecture=_platform.architecture()[0],
        version=_platform.version()):
    """
        Find supported run time id for current platform.
    """
    run_time_id = None

    if system == 'Windows':
        if architecture == '32bit':
            run_time_id = 'Windows_7_86'
        elif architecture == '64bit':
            run_time_id = 'Windows_7_64'
    elif system == 'Darwin':
        if architecture == '64bit':
            run_time_id = 'OSX_10_11_64'
    elif system == 'Linux':
        if architecture == '64bit':
            run_time_id = _get_linux_distro_from_file()

    return run_time_id


def get_mssqltoolsservice_package_name(run_time_id=_get_runtime_id()):
    """
        Retrieve sql tools service package name for this platform if supported.
    """
    if run_time_id and run_time_id in MSSQLTOOLSSERVICE_PACKAGE_SUFFIX:
        # set package suffix name for other uses like building wheels outside of setup.py.
        os.environ['MSSQLTOOLSSERVICE_PACKAGE_SUFFIX'] = run_time_id
        return MSSQLTOOLSSERVICE_PACKAGE_NAME.format(
            run_time_id, MSSQLSCRIPTER_VERSION).replace('_', '-').lower()

    raise EnvironmentError('mssqltoolsservice is not supported on this platform.')


CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Intended Audience :: System Administrators',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'License :: OSI Approved :: MIT License',
]

DEPENDENCIES = [
    'future>=0.16.0',
    'wheel>=0.29.0'
]

if sys.version_info < (3, 4):
    DEPENDENCIES.append('enum34>=1.1.6')

DEPENDENCIES.append(get_mssqltoolsservice_package_name())

# Using a environment variable to communicate mssqltoolsservice package name for
# other modules that need that info like dev_setup.py.
os.environ['MSSQLTOOLSSERVICE_PACKAGE_NAME'] = DEPENDENCIES[-1]

setup(
    install_requires=DEPENDENCIES,
    name='mssql-scripter',
    version=MSSQLSCRIPTER_VERSION,
    description='Microsoft SQL Scripter Command-Line Tool',
    license='MIT',
    author='Microsoft Corporation',
    author_email='sqlxplatclieng@microsoft.com',
    url='https://github.com/Microsoft/sql-xplat-cli/',
    zip_safe=True,
    long_description=open('README.rst').read(),
    classifiers=CLASSIFIERS,
    include_package_data=True,
    scripts=[
        'mssql-scripter',
        'mssql-scripter.bat'
    ],
    packages=[
        'mssqlscripter',
        'mssqlscripter.jsonrpc',
        'mssqlscripter.jsonrpc.contracts'],
)
